using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using Singleton.CGraphics;
using Singleton.CProteinManager;
using Singleton.CCellManager;
using Singleton.CSound;

namespace CWorld
{
    public class CCell : CWorldEntity
    {
        #region fields and their properties

        private Vector2 entityDirection;

        public Vector2 EntityDirection
        {
            get { return entityDirection; }
            set { entityDirection = value; }
        }

        private Vector2 targetDirection;

        public Vector2 TargetDirection
        {
            get { return targetDirection; }
            set { targetDirection = value; }
        }

        private int entityVision;

        public int EntityVision
        {
            get { return entityVision; }
            set { entityVision = value; }
        }

        private int repulsionRange;

        public int RepulsionRange
        {
            get { return repulsionRange; }
            set { repulsionRange = value; }
        }

        private int currentTargetID;

        public int CurrentTargetID
        {
            get { return currentTargetID; }
            set { currentTargetID = value; }
        }
        private float infectionLevel;

        public float InfectionLevel
        {
            get { return infectionLevel; }
            set { infectionLevel = value; }
        }
        private float infectionRate;

        public float InfectionRate
        {
            get { return infectionRate; }
            set { infectionRate = value; }
        }

        private float begunInfectedTime;

        public float BegunInfectedTime
        {
            get { return begunInfectedTime; }
            set { begunInfectedTime = value; }
        }

        private float DEFAULT_INFECTED_TIME;

        public float pDEFAULT_INFECTED_TIME
        {
            get { return DEFAULT_INFECTED_TIME; }
        }

        private ICell cellBehaviour;

        public ICell CellBehaviour
        {
            get { return cellBehaviour; }
            set { cellBehaviour = value; }
        }

        private int experienceGiven;

        public int ExperienceGiven
        {
            get { return experienceGiven; }
            set { experienceGiven = value; }
        }

        private int numberOfViruses;

        public int NumberOfViruses
        {
            get { return numberOfViruses; }
            set { numberOfViruses = value; }
        }

        private Color infectedColour = Color.Purple;

        public Color InfectedColour
        {
            get { return infectedColour; }
            set { infectedColour = value; }
        }
        #endregion

        #region constructor
        public CCell(int newGraphicID, int newHealth, int newSpeed, int newPointValue, int newCollisionValue,
            Vector2 newPosition, int newWidth, int newHeight, int newVision, float newInfectionRate, float newInfectedTime,
            int newExperienceGiven) 
            : base(newGraphicID, newHealth, newSpeed, newPointValue, newCollisionValue, newPosition, newWidth, newHeight)
        {
            entitySide = eSide.cell;

            // intialising the variables relating to finite state machine
            entityDirection = new Vector2(1, 1);
            entityVision = newVision;
            repulsionRange = 50;
            currentTargetID = -1;
            infectionLevel = 0.0f;
            infectionRate = newInfectionRate;
            DEFAULT_INFECTED_TIME = newInfectedTime;
            begunInfectedTime = 0;
            cellBehaviour = new CRoamingCell();
            experienceGiven = newExperienceGiven;
            pointValue = newPointValue;
        }
        #endregion

        #region methods
        public override void update()
        {
            if (isActive)
            {
                // updates the previous position
                previousPosition = currentPosition;

                // updates the collision rectangle
                updateCollisionRectangle(currentPosition);

                // normalises the direction before using it for movement
                if (entityDirection != Vector2.Zero)
                {
                    entityDirection.Normalize();
                }

                // advances the entity in the current direction times the speed
                currentPosition += (entityDirection * entitySpeed);

                // resets the direction
                entityDirection = Vector2.Zero;
            }
        }
        public override void renderEntity(float interpolateFract)
        {
            if (isActive == true)
            {
                Trace.Assert((interpolateFract >= 0) && (interpolateFract <= 1), "Interpolation value out of bounds."
                       , "The interpolate value is currently " + interpolateFract.ToString());

                Vector2 interVector = previousPosition;

                if (isActive)
                {
                    // only interpolates when necessary
                    if (previousPosition != currentPosition)
                    {
                        interVector += interpolateFract * (currentPosition - previousPosition);
                    }
                }

                CGraphics.get.DrawSprite(graphicID, (int)interVector.X, (int)interVector.Y, infectedColour, infectionLevel);
            }
        }

        public override void takeDamage(int damage)
        {
            if (invincible == false)
            {
                entityHealth -= damage;
                if (entityHealth <= 0)
                {
                    entityHealth = 0;
                    isActive = false;
                    attackable = false;
                    CSound.getInstance.PlaySoundEffect(CWorld.Game1.deathSoundId);
                }
            }
        }

        // takes in the side of another entity and checks to see if it is it's enemy
        public override bool isEnemyOf(eSide otherSide)
        {
            // checks to see if the other entity is an enemy
            return (otherSide == eSide.player || otherSide == eSide.virus || otherSide == eSide.projectile);
        }

        // a function to determine whether another position is within this entity's
        // pickUp of vision
        public bool inRange(Vector2 otherPosition, int rangeCheck)
        {
            return (distanceBetween(otherPosition) <= rangeCheck);
        }

        // a function to set the direction of the current entity
        // pass in the position to aim the entity towards
        public void determineDirection(Vector2 otherPosition)
        {
            entityDirection += otherPosition - currentPosition;
        }

        // a function to avoid targets
        public Vector2 avoid()
        {
            // creates a temporary vector 
            Vector2 temp = Vector2.Zero;

            for (int i = CCellManager.getInstance.FirstCell; i < CCellManager.getInstance.FirstCell + CCellManager.getInstance.CellCount; i++)
            {
                // ensures the entity is not comparing against itself
                if (CWorld.Game1.entityList[i] != this)
                {
                    // checks to see if the other entity is active
                    if (CWorld.Game1.entityList[i].IsActive)
                    {
                        // checks to see if the entity is range
                        if (inRange(CWorld.Game1.entityList[i].CurrentPosition, repulsionRange))
                        {
                            temp += currentPosition - CWorld.Game1.entityList[i].CurrentPosition;
                        }
                    }
                }
            }

            // returns the suggested avoidance vector
            return temp;
        }


        // determines the action of the cell based on it's current state
        public override void determineAction(GameTime time)
        {
            cellBehaviour.determineAction(this, time);
        }

        // the result of the collision function
        public override void collisionResult(int otherEntityID)
        {
            cellBehaviour.collisionResult(this, otherEntityID);
        }
        #endregion
    }
}